﻿using System;
using System.Collections.Generic;
using JetBrains.Annotations;
using UniRx;
using System.Threading;
using Cysharp.Threading.Tasks;

namespace Magify
{
    internal interface ILegacyBridgePlatformAPI
    {
        event Action<LtoInfo> OnLtoDidFinish;
        event Action<IReadOnlyCollection<LtoInfo>> OnLtoDidUpdate;
    }

    internal interface IModernBridgePlatformAPI
    {
        event Action<LtoInfo> OnOfferAdded;
        event Action<LtoInfo> OnOfferUpdated;
        event Action<LtoInfo> OnOfferRemoved;
        event Action<LtoInfo> OnOfferFinished;
    }

    internal abstract class MagifyPlatformAPI : IDisposable
    {
        internal abstract void CompleteOffer(string offerName);
        internal abstract IReadOnlyCollection<LtoInfo> ActiveLtoOffers { get; }

        internal abstract event Action OnConfigLoaded;

        internal abstract Environment Environment { get; set; }
        [NotNull]
        internal abstract IReactiveProperty<IPurchaseVerificationHandler> ExternalPurchaseVerificationHandler { get; }
        [NotNull]
        internal abstract FloatReactiveProperty VerificationRetryInterval { get; }
        internal abstract SubscriptionStatus SubscriptionStatus { get; set; }
        internal abstract bool IsLoggingEnabled { get; set; }
        internal abstract string ReferrerId { get; set; }
        internal abstract string AdjustId { get; set; }
        internal abstract string FirebaseInstanceId { get; set; }
        internal abstract bool IsPortrait { get; }
        internal abstract InAppStatus InAppStatus { get; set; }

        internal abstract ContextSyncTime LastContextSyncTime { get; }

        /// <remarks>
        /// During a Reset instance of the Features might be changed.
        /// </remarks>
        internal abstract IFeatures Features { get; }

        internal abstract Limits Limits { get; }
        internal abstract string ClientId { get; }

        internal abstract IReactiveProperty<TimeSpan> SessionsInterval { get; }
        internal abstract int SessionNumber { get; }
        internal abstract string FirstInstalledVersion { get; }
        [NotNull]
        internal abstract IReactiveProperty<ConfigScope> RemoteContextScope { get; }
        internal abstract event Action<int> OnSessionChanged;

        internal abstract event Action OnSubscriptionStatusChanged;
        internal abstract event Action OnInAppStatusChanged;
        internal abstract event Action OnAuthorizationStatusChanged;
        internal abstract event Action OnReferrerChanged;
        internal abstract event Action OnPurchasedProductsChanged;

        internal abstract event Action OnApplicationEnterForeground;
        internal abstract event Action OnApplicationEnterBackground;

        internal abstract event Action OnUserDidTakeScreenshot;

        internal abstract AuthorizationStatus AuthorizationStatus { get; set; }

        internal abstract List<InAppProduct> InAppProducts { get; }
        internal abstract List<SubscriptionProduct> SubscriptionProducts { get; }
        [NotNull]
        internal abstract IReactiveProperty<bool> SyncStateEnabled { get; }
        [NotNull]
        internal abstract IObservable<SyncStateResult> OnRestoreStateCompleted { get; }

        internal abstract bool IsSdkInitialized { get; }
        internal abstract IReadOnlyReactiveProperty<IReadOnlyList<string>> Segmentations { get; }
        internal abstract IReadOnlyReactiveProperty<IReadOnlyList<AssignedAbTest>> AssignedAbTests { get; }
        [NotNull]
        internal abstract IStoredAppFeaturesCollection StoredAppFeatures { get; }

        internal abstract void InitializeSdk(string appName, string configPath, string clientId, Environment configEnvironment);

        internal abstract void Reset(bool clearNativeStorage = true, bool clearCloudStorage = true);
        internal abstract void ClearNativeStorage();
        internal abstract void Setup(TaskScheduler taskScheduler);
        internal abstract void Sync();
        internal abstract void ResetAndForceUpdate();

        internal abstract void TweakAnalyticsConfig(int eventsGroupSize, int syncTimeInterval);
        internal abstract void ResetAnalyticsConfig();
        internal abstract void TweakUserLocale(string languageTag);

        internal abstract void SetMediaSource(string networkName = null, string campaignName = null, string adGroup = null);

        internal abstract void SetAttStatus(bool authorized);
        internal abstract void SetupConversionTracker(string revenuePerCountryPath, string revenueLevelsPath, string defaultCurrencyRatesPath, string subscriptionMultipliersPath);

        internal abstract ICampaign CampaignFor(string eventName, IReadOnlyDictionary<string, object> customParams = null, bool silent = false);
        internal abstract CampaignImpression LastImpressionFor(CampaignType campaignType, string campaignName, string eventName);
        internal abstract bool IsCampaignAvailable(string campaignName, string eventName, Dictionary<string, object> customParams = null);

        internal abstract void SubscribeCampaignUpdates(string campaignName, Action<ICampaign> onUpdate);
        internal abstract void UnsubscribeCampaignUpdates(string campaignName);

        internal abstract bool HasProcessedPurchase(string productId);

        internal abstract void SetGameMode(string mode);
        internal abstract void SetGameLevel(int level);
        internal abstract void SetGameMaxLevel(int maxLevel);

        internal abstract void SetUserEmail(string email);

        internal abstract UniTask<bool> RequestSocialAuthTokenFor(string provider, string token, CancellationToken cancellationToken);
        internal abstract void ResetSocialAuthToken();
        internal abstract UniTask<SyncStateResult> SaveState(int? weight, CancellationToken cancellationToken);
        internal abstract UniTask<SyncStateResult> RestoreState(int? weight, CancellationToken cancellationToken);

        internal abstract void TrackGdprAccessState(bool accessState);

        internal abstract void TrackAppLaunch();

        internal abstract void TrackCustomEvent(string eventName, IReadOnlyDictionary<string, object> customParams = null);

        internal abstract void TrackImpression(CampaignType campaignType);
        internal abstract void TrackParentCampaignImpression(CampaignType campaignType);
        internal abstract void TrackProductsImpression(CampaignType campaignType, List<string> productIds);

        internal abstract void TrackAdsImpression(CampaignType campaignType, IAdsImpression impression);
        internal abstract void TrackImpressionFailFor(CampaignType campaignType, string reason);

        internal abstract void TrackClickFor(CampaignType campaignType);
        internal abstract void TrackAdsClickFor(CampaignType campaignType);
        internal abstract void TrackProductClickFor(CampaignType campaignType, string productId);
        internal abstract void TrackAdsProductClickFor(CampaignType campaignType, string productId);

        internal abstract void TrackSubscriptionActivation(bool isTrial, string productId, string price, string currency, string period, string transactionId, string purchaseToken, string originalTransactionId, string receipt);
        internal abstract void TrackExternalSubscriptionActivation(bool isTrial, string productId, string price, string currency, string transactionId, string purchaseToken, string originalTransactionId, string receipt);
        internal abstract void TrackInAppFor(string productId, string price, string currency, string transactionId, string purchaseToken, string originalTransactionId, string receipt);
        internal abstract void TrackExternalInAppFor(string productId, string price, string currency, string transactionId, string purchaseToken, string originalTransactionId, string receipt);
        internal abstract void TrackRestoredInAppFor(string productId);
        internal abstract void TrackRestoredSubscription(string productId);

        internal abstract void TrackIncomeTransaction(string source, List<BonusInfo> bonuses, ProductInfo product = null);
        internal abstract void TrackExpenseTransaction(List<BonusInfo> bonuses);
        internal abstract void TrackCorrectionTransaction(List<BonusInfo> bonuses);

        internal abstract void TrackRewardGranted(string productId);
        internal abstract void TrackFreeBonusGranted(string productId);
        internal abstract void TrackOrdinaryProductUsed(string productId);

        internal abstract List<ContentItem> GetContentList(string group, string key, List<string> tags = null);
        internal abstract ContentItem GetEarliestContent(string group, string key);
        internal abstract ContentItem GetLatestContent(string group, string key);

        public abstract void Dispose();
    }
}