﻿using System;
using System.Collections.Generic;
using System.Threading;
using Cysharp.Threading.Tasks;
using JetBrains.Annotations;
using Newtonsoft.Json;
using Newtonsoft.Json.Utilities;
using UniRx;
using UnityEngine;
using Object = UnityEngine.Object;

namespace Magify
{
    public static partial class MagifyManager
    {
        public static event Action OnConfigLoaded;

        public static event Action<int> OnSessionChanged;

        public static event Action OnSubscriptionStatusChanged;
        public static event Action OnInAppStatusChanged;
        public static event Action OnAuthorizationStatusChanged;
        public static event Action OnReferrerChanged;
        public static event Action OnPurchasedProductsChanged;

        public static event Action OnApplicationEnterForeground;
        public static event Action OnApplicationEnterBackground;

        /// <summary>
        /// Only supported on iOS
        /// </summary>
        public static event Action OnUserDidTakeScreenshot;

        private static MagifyLogger _logger;
        private static Transform _root;
        private static MagifyPresenter _presenter;

        /// <remarks>
        /// Null can only be set at ShutDown
        /// </remarks>
        internal static MagifyPlatformAPI MagifyPlatformAPI { get; private set; }

        public static bool Initialized => MagifyPlatformAPI is { IsSdkInitialized: true };

        /// <inheritdoc cref="ThrowIfMagifyIsNotReady"/>
        /// <summary>
        /// Allows you to toggle application state synchronization.
        /// </summary>
        /// <remarks>
        /// If you want to control this property since initialization you can use <see cref="MagifyConfig.SyncStateEnabled"/>. <br/>
        /// If this flag was set to 'true' since initialization, the application state will attempt to be restored automatically. <br/>
        /// In case the flag was set to 'true' after initialization, no automatic restore will take place.
        /// </remarks>
        [NotNull]
        public static IReactiveProperty<bool> SyncStateEnabled
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(SyncStateEnabled));
                return MagifyPlatformAPI.SyncStateEnabled;
            }
        }

        /// <inheritdoc cref="ThrowIfMagifyIsNotReady"/>
        /// <summary>
        /// Allows you to handle successful state synchronization.
        /// </summary>
        /// <returns>
        /// <see cref="SyncStateResult"/>, that indicates restore result <br/>
        /// <see cref="SyncStateResult"/>.<see cref="SyncStateResult.Kind"/> might be only
        /// <see cref="SyncStateResult.ResultKind.Success"/> or <see cref="SyncStateResult.ResultKind.NothingToRestore"/> here.
        /// </returns>
        /// <remarks>
        /// If you want to have more control over the result of the restore, use the return value <see cref="RestoreState"/>.
        /// </remarks>
        [NotNull]
        public static IObservable<SyncStateResult> OnRestoreStateCompleted
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(OnRestoreStateCompleted));
                return MagifyPlatformAPI.OnRestoreStateCompleted;
            }
        }

        public static ContextSyncTime LastContextSyncTime
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(LastContextSyncTime));
                return MagifyPlatformAPI.LastContextSyncTime;
            }
        }

        public static bool IsPortrait
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(IsPortrait));
                return MagifyPlatformAPI.IsPortrait;
            }
        }

        public static Environment Environment
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(Environment));
                return MagifyPlatformAPI.Environment;
            }
            set
            {
                ThrowIfMagifyIsNotReady(nameof(Environment));
                _logger.Log($"Set {nameof(Environment)} to {value}");
                MagifyPlatformAPI.Environment = value;
            }
        }

        /// <summary>
        /// Allows you to control purchases verification process and get the result of it
        /// </summary>
        [NotNull]
        public static IReactiveProperty<IPurchaseVerificationHandler> ExternalPurchaseVerificationHandler
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(ExternalPurchaseVerificationHandler));
                return MagifyPlatformAPI.ExternalPurchaseVerificationHandler;
            }
        }

        /// <summary>
        /// The time in seconds between purchase verification attempts.
        /// </summary>
        /// <remarks>
        /// Minimum allowed value is 1 second.
        /// </remarks>
        [NotNull]
        public static FloatReactiveProperty VerificationRetryInterval
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(VerificationRetryInterval));
                return MagifyPlatformAPI.VerificationRetryInterval;
            }
        }

        public static SubscriptionStatus SubscriptionStatus
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(SubscriptionStatus));
                return MagifyPlatformAPI.SubscriptionStatus;
            }
            set
            {
                ThrowIfMagifyIsNotReady(nameof(SubscriptionStatus));
                _logger.Log($"Set {nameof(SubscriptionStatus)} to {value}");
                MagifyPlatformAPI.SubscriptionStatus = value;
            }
        }

        public static AuthorizationStatus AuthorizationStatus
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(AuthorizationStatus));
                return MagifyPlatformAPI.AuthorizationStatus;
            }
            set
            {
                ThrowIfMagifyIsNotReady(nameof(AuthorizationStatus));
                _logger.Log($"Set {nameof(AuthorizationStatus)} to {value}");
                MagifyPlatformAPI.AuthorizationStatus = value;
            }
        }

        public static InAppStatus InAppStatus
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(InAppStatus));
                return MagifyPlatformAPI.InAppStatus;
            }
            set
            {
                ThrowIfMagifyIsNotReady(nameof(InAppStatus));
                MagifyPlatformAPI.InAppStatus = value;
            }
        }

        public static string ReferrerId
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(ReferrerId));
                return MagifyPlatformAPI.ReferrerId;
            }
            set
            {
                ThrowIfMagifyIsNotReady(nameof(ReferrerId));
                _logger.Log($"Set {nameof(ReferrerId)} to {value}");
                MagifyPlatformAPI.ReferrerId = value;
            }
        }

        public static string AdjustId
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(AdjustId));
                return MagifyPlatformAPI.AdjustId;
            }
            set
            {
                ThrowIfMagifyIsNotReady(nameof(AdjustId));
                _logger.Log($"Set {nameof(AdjustId)} to {value}");
                MagifyPlatformAPI.AdjustId = value;
            }
        }

        public static string FirebaseInstanceId
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(FirebaseInstanceId));
                return MagifyPlatformAPI.FirebaseInstanceId;
            }
            set
            {
                ThrowIfMagifyIsNotReady(nameof(FirebaseInstanceId));
                _logger.Log($"Set {nameof(FirebaseInstanceId)} to {value}");
                MagifyPlatformAPI.FirebaseInstanceId = value;
            }
        }

        public static Limits Limits
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(Limits));
                return MagifyPlatformAPI.Limits;
            }
        }

        public static string ClientId
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(ClientId));
                return MagifyPlatformAPI.ClientId;
            }
        }

        public static List<InAppProduct> InAppProducts
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(InAppProducts));
                return MagifyPlatformAPI.InAppProducts;
            }
        }

        public static List<SubscriptionProduct> SubscriptionProducts
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(SubscriptionProducts));
                return MagifyPlatformAPI.SubscriptionProducts;
            }
        }

        /// <summary>
        /// Returns a list of segments for which the user is suitable.
        /// </summary>
        /// <remarks>
        /// Might be received only from remote config. There is no in default config.
        /// </remarks>
        [NotNull]
        public static IReadOnlyReactiveProperty<IReadOnlyList<string>> Segmentations
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(Segmentations));
                return MagifyPlatformAPI.Segmentations;
            }
        }

        /// <summary>
        /// Returns a list of a/b-tests and groups for which the user is suitable.
        /// </summary>
        /// <remarks>
        /// Might be received only from remote config. There is no in default config.
        /// </remarks>
        [NotNull]
        public static IReadOnlyReactiveProperty<IReadOnlyList<AssignedAbTest>> AssignedAbTests
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(AssignedAbTests));
                return MagifyPlatformAPI.AssignedAbTests;
            }
        }

        /// <summary>
        /// Be careful with this property, in the Full version of Magify Sdk it will be loaded with app config/context.
        /// </summary>
        public static IReactiveProperty<TimeSpan> SessionsInterval
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(SessionsInterval));
                return MagifyPlatformAPI.SessionsInterval;
            }
        }

        public static int SessionNumber
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(SessionNumber));
                return MagifyPlatformAPI.SessionNumber;
            }
        }

        /// <summary>
        /// Returns first installed version of the app (that had Magify SDK integrated)
        /// </summary>
        public static string FirstInstalledVersion
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(FirstInstalledVersion));
                return MagifyPlatformAPI.FirstInstalledVersion;
            }
        }

        /// <inheritdoc cref="ThrowIfMagifyIsNotReady"/>
        /// <inheritdoc cref="ConfigScope"/>
        [NotNull]
        public static IReactiveProperty<ConfigScope> RemoteConfigScopes
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(RemoteConfigScopes));
                return MagifyPlatformAPI.RemoteContextScope;
            }
        }

        public static MagifyPresenter Presenter
        {
            get
            {
                ThrowIfMagifyIsNotReady(nameof(Presenter));
                return _presenter;
            }
        }

        public static Transform Root => _root;

        public static void Initialize(MagifyConfig config, [CanBeNull] MagifyDebugConfig debugConfig = null)
        {
            AotHelper.EnsureList<LtoModel>();
            AotHelper.EnsureDictionary<string, object>();

            if (MagifyPlatformAPI != null)
            {
                throw new MagifyAlreadyInitializedException();
            }

            _logger = MagifyLogger.Get();
            MagifyLogger.IsLoggingEnabled = config.IsLoggingEnabled;

            var editorDevice = config.EditorDevice ?? EditorDevice.GetDefaultAndroidDevice();
            var platform = GetCurrentPlatform(editorDevice);
            var appName = platform switch
            {
                RuntimePlatform.Android => config.AppNameGP,
                RuntimePlatform.IPhonePlayer => config.AppNameIOS,
                _ => string.Empty
            };
            var configPath = platform switch
            {
                RuntimePlatform.Android => config.ConfigPathGP,
                RuntimePlatform.IPhonePlayer => config.ConfigPathIOS,
                _ => string.Empty
            };

            if (string.IsNullOrEmpty(appName))
            {
                throw new NoAppNameForMagifyInitializationException(platform);
            }

            MagifyPlatformAPI = new MagifyCSharp(config.EditorMoq, config.IsDeveloperMode, !config.UseAdvancedVersion, editorDevice, appName, configPath, config.StoragePath, config.ClientStateConfig, config.Environment);

#if MAGIFY_VERBOSE_LOGGING
            try
            {
                if (debugConfig != null)
                {
                    (MagifyPlatformAPI as MagifyCSharp)!.Client!.Platform!.CustomAndroidId = debugConfig.CustomAndroidId;
                    (MagifyPlatformAPI as MagifyCSharp)!.Client!.Platform!.CustomBuildNumber = debugConfig.CustomBuildNumber;
                }
            }
            catch (Exception){ /* Ignore */ }
#endif

            _logger.Log("Create root GameObject");
            var gameObject = new GameObject("Magify SDK");
            Object.DontDestroyOnLoad(gameObject);
            _root = gameObject.transform;
            Root.position = Vector3.zero;
            Root.rotation = Quaternion.identity;
            Root.localScale = Vector3.one;
            var taskScheduler = gameObject.AddComponent<TaskScheduler>();

            _logger.Log("Initialize LTO");
            Lto.Initialize(taskScheduler);
            _logger.Log("Initialize Features");
            Features.Initialize();

            // Specific for migration from iOS native SDK
            // ToDo: remove after removing of migration from native iOS SDK
            SubscriptionStatus = config.SubscriptionStatus;
            AuthorizationStatus = config.AuthorizationStatus;

            if (config.SyncStateEnabled != null)
            {
                MagifyPlatformAPI.SyncStateEnabled.Value = config.SyncStateEnabled.Value;
            }
            MagifyPlatformAPI.RemoteContextScope.Value = config.RemoteConfigScopes;

            _logger.Log($"Initialize Magify SDK ({MagifyPlatformAPI.GetType().Name}). {nameof(appName)}={appName}; {nameof(configPath)}={configPath}");
            MagifyPlatformAPI.InitializeSdk(appName, configPath, config.CustomClientId, config.Environment);
            _logger.Log($"Magify SDK has been initialized. {nameof(ClientId)} is {ClientId}");
            MagifyPlatformAPI.IsLoggingEnabled = config.IsLoggingEnabled;

            Environment = config.Environment;
            SubscriptionStatus = config.SubscriptionStatus;
            AuthorizationStatus = config.AuthorizationStatus;
            if (config.IsGdprApplied != null)
            {
                TrackGdprAccessState(config.IsGdprApplied.Value);
            }
            if (config.IsAttAuthorized != null)
            {
                SetAttStatus(config.IsAttAuthorized.Value);
            }

            MagifyPlatformAPI.SetupConversionTracker(
                $"{Application.streamingAssetsPath}/{config.RevenuePerCountryPath}",
                $"{Application.streamingAssetsPath}/{config.RevenueLevelsPath}",
                $"{Application.streamingAssetsPath}/{config.DefaultCurrencyRatesPath}",
                $"{Application.streamingAssetsPath}/{config.SubscriptionMultipliersPath}"
            );

            _logger.Log("Setup SDK");
            MagifyPlatformAPI.Setup(taskScheduler);
            _logger.Log("Prepare storages");
            Storage.Setup(config.StoragePath);
            _logger.Log("Setup lto offers");
            Lto.Setup();
            _logger.Log("Init native callbacks");
            InitCallbacks();
            _logger.Log("Prepare presenter");
            _presenter = MagifyPresenter.BuildDefaultPresenter(config.PresenterSettings, Root);
        }

        private static RuntimePlatform GetCurrentPlatform(EditorDevice editorDevice)
        {
            if (!Application.isEditor)
            {
                return Application.platform;
            }
            return editorDevice?.Platform ?? Application.platform;
        }

        private static void InitCallbacks()
        {
            MagifyPlatformAPI.OnConfigLoaded += () =>
            {
                _logger.Log($"{nameof(OnConfigLoaded)} has been called.");
                OnConfigLoaded?.Invoke();
            };

            MagifyPlatformAPI.OnSessionChanged += (sessionNumber) =>
            {
                _logger.Log($"{nameof(OnSessionChanged)} has been called.");
                OnSessionChanged?.Invoke(sessionNumber);
            };

            MagifyPlatformAPI.OnSubscriptionStatusChanged += () =>
            {
                _logger.Log($"{nameof(OnSubscriptionStatusChanged)} has been called.");
                OnSubscriptionStatusChanged?.Invoke();
            };

            MagifyPlatformAPI.OnInAppStatusChanged += () =>
            {
                _logger.Log($"{nameof(OnInAppStatusChanged)} has been called.");
                OnInAppStatusChanged?.Invoke();
            };

            MagifyPlatformAPI.OnAuthorizationStatusChanged += () =>
            {
                _logger.Log($"{nameof(OnAuthorizationStatusChanged)} has been called.");
                OnAuthorizationStatusChanged?.Invoke();
            };

            MagifyPlatformAPI.OnReferrerChanged += () =>
            {
                _logger.Log($"{nameof(OnReferrerChanged)} has been called.");
                OnReferrerChanged?.Invoke();
            };

            MagifyPlatformAPI.OnPurchasedProductsChanged += () =>
            {
                _logger.Log($"{nameof(OnPurchasedProductsChanged)} has been called.");
                OnPurchasedProductsChanged?.Invoke();
            };

            MagifyPlatformAPI.OnApplicationEnterForeground += () =>
            {
                _logger.Log($"{nameof(OnApplicationEnterForeground)} has been called.");
                OnApplicationEnterForeground?.Invoke();
            };

            MagifyPlatformAPI.OnApplicationEnterBackground += () =>
            {
                _logger.Log($"{nameof(OnApplicationEnterBackground)} has been called.");
                OnApplicationEnterBackground?.Invoke();
            };

            MagifyPlatformAPI.OnUserDidTakeScreenshot += () =>
            {
                _logger.Log($"{nameof(OnUserDidTakeScreenshot)} has been called.");
                OnUserDidTakeScreenshot?.Invoke();
            };
        }

        public static void ShutDown()
        {
            if (MagifyPlatformAPI != null)
            {
                MagifyPlatformAPI.Dispose();
                MagifyPlatformAPI = default;
            }

            OnConfigLoaded = null;
            OnSessionChanged = null;
            OnSubscriptionStatusChanged = null;
            OnInAppStatusChanged = null;
            OnAuthorizationStatusChanged = null;
            OnReferrerChanged = null;
            OnPurchasedProductsChanged = null;
            OnApplicationEnterForeground = null;
            OnApplicationEnterBackground = null;
            OnUserDidTakeScreenshot = null;

            Lto.Reset();
            Features.Reset();
            Storage.Clear();
            Logging.Reset();

            if (_presenter != null) destroy(_presenter.gameObject);
            if (_root != null) destroy(_root.gameObject);
            _root = default;
            _presenter = default;
            _logger = default;

            void destroy(Object obj)
            {
                switch (obj != null, Application.isPlaying)
                {
                    case (true, true):
                        Object.Destroy(obj);
                        break;
                    case (true, false):
                        Object.DestroyImmediate(obj);
                        break;
                }
            }
        }

        public static void Sync()
        {
            ThrowIfMagifyIsNotReady(nameof(Sync));
            _logger.Log("Sync config with backend");
            MagifyPlatformAPI.Sync();
        }

        public static void Reset(bool clearNativeStorage = true, bool clearCloudStorage = true)
        {
            ThrowIfMagifyIsNotReady(nameof(Reset));

            OnConfigLoaded = null;
            OnSessionChanged = null;
            OnSubscriptionStatusChanged = null;
            OnInAppStatusChanged = null;
            OnAuthorizationStatusChanged = null;
            OnReferrerChanged = null;
            OnPurchasedProductsChanged = null;
            OnApplicationEnterForeground = null;
            OnApplicationEnterBackground = null;
            OnUserDidTakeScreenshot = null;
            Lto.Reset();
            Features.Reset();
            Storage.Clear();
            Logging.Reset();
            MagifyPlatformAPI.Reset(clearNativeStorage, clearCloudStorage);
        }

        /// <inheritdoc cref="MagifyCSharp.ClearNativeStorage"/>
        public static void ClearNativeStorage()
        {
            MagifyPlatformAPI.ClearNativeStorage();
        }

        public static void ResetAndForceUpdate()
        {
            ThrowIfMagifyIsNotReady(nameof(ResetAndForceUpdate));
            MagifyPlatformAPI.ResetAndForceUpdate();
        }

        public static void TweakAnalyticsConfig(int eventsGroupSize, int syncTimeInterval)
        {
            ThrowIfMagifyIsNotReady(nameof(TweakAnalyticsConfig));
            MagifyPlatformAPI.TweakAnalyticsConfig(eventsGroupSize, syncTimeInterval);
        }

        public static void ResetAnalyticsConfig()
        {
            ThrowIfMagifyIsNotReady(nameof(ResetAnalyticsConfig));
            MagifyPlatformAPI.ResetAnalyticsConfig();
        }

        public static void TweakUserLocale(string languageTag)
        {
            ThrowIfMagifyIsNotReady(nameof(TweakUserLocale));
            MagifyPlatformAPI.TweakUserLocale(languageTag);
        }

        public static void SetAttStatus(bool authorized)
        {
            ThrowIfMagifyIsNotReady(nameof(SetAttStatus));
            MagifyPlatformAPI.SetAttStatus(authorized);
        }

        public static void SetMediaSource(string networkName = null, string campaignName = null, string adGroup = null)
        {
            ThrowIfMagifyIsNotReady(nameof(SetMediaSource));
            MagifyPlatformAPI.SetMediaSource(networkName, campaignName, adGroup);
        }

        public static void SetUserEmail(string email)
        {
            ThrowIfMagifyIsNotReady(nameof(SetUserEmail));
            if (EmailValidator.IsValidEmail(email, out var error))
            {
                MagifyPlatformAPI.SetUserEmail(email);
            }
            else
            {
                _logger.LogError($"The user's email has an invalid format, validation failed with an error: {error}");
            }
        }

        /// <inheritdoc cref="ThrowIfMagifyIsNotReady"/>
        /// <summary>
        /// Requests authorization with the social authorization data.
        /// </summary>
        /// <param name="provider">The name of the social authorization method. You can select it yourself depends on your environment.</param>
        /// <param name="token">A unique token for each user that will be associated with the current user account.</param>
        /// <param name="cancellationToken">Pass cancellation token if you want to cancel this operation, for example by timeout.</param>
        /// <returns>
        /// Result of the authorization operation. True if successful.
        /// </returns>
        /// <example>
        /// <code>
        /// AuthorizeUser(provider: "google", token: "3A2u6a6fZoxHP7n0AosZbv3k8zvy0Y", cancellationToken);
        /// </code>
        /// </example>
        public static UniTask<bool> AuthorizeUser([NotNull] string provider, [NotNull] string token, CancellationToken cancellationToken)
        {
            ThrowIfMagifyIsNotReady(nameof(AuthorizeUser));
            if (string.IsNullOrEmpty(provider) || string.IsNullOrEmpty(token))
            {
                _logger.LogWarning($"Parameters {nameof(provider)} and {nameof(token)} cannot be null or empty, skipping");
                return UniTask.FromResult(false);
            }
            _logger.Log($"Authorize user with {nameof(provider)}={provider}; {nameof(token)}={token}");
            return MagifyPlatformAPI.RequestSocialAuthTokenFor(provider, token, cancellationToken);
        }

        /// <inheritdoc cref="ThrowIfMagifyIsNotReady"/>
        /// <summary>
        /// Resets the authorization data (provider and token) that were set earlier via <see cref="AuthorizeUser"/>.
        /// From now, saving and restoring progress will work on the basis of `ClientId` only, until next <see cref="AuthorizeUser"/> call.
        /// </summary>
        public static void ResetUserAuthorization()
        {
            ThrowIfMagifyIsNotReady(nameof(ResetUserAuthorization));
            _logger.Log("Reset user authorization");
            MagifyPlatformAPI.ResetSocialAuthToken();
        }

        /// <inheritdoc cref="ThrowIfMagifyIsNotReady"/>
        /// <summary>
        /// Requests saving the state of the application. You shouldn't call more than one operation in one time.
        /// </summary>
        /// <param name="weight"> A weight that determines how far a player has gone in the game.
        ///     <ul>
        ///         <li> If the weight parameter is specified:
        ///             <ul>
        ///                 <li>if the value <b>is greater than</b> the value on the server - progress <b>will</b> be saved</li>
        ///                 <li>if the value <b>is less than</b> the value on the server - progress <b>won’t</b> be saved</li>
        ///             </ul>
        ///         </li>
        ///         <li> If the weight parameter is not specified:
        ///             <ul>
        ///                 <li>if <b>there is any</b> value on the server - progress <b>won’t</b> be saved</li>
        ///             </ul>
        ///         </li>
        ///         <li> In the both cases:
        ///             <ul>
        ///                 <li>if there is no value (null) on the server - progress will be saved</li>
        ///             </ul>
        ///         </li>
        ///     </ul>
        ///  <br/>
        /// </param>
        /// <param name="cancellationToken">Pass cancellation token if you want to cancel this operation, for example by timeout.</param>
        /// <returns>
        /// Result of the save operation.
        /// </returns>
        /// <remarks>
        /// Remember that you should pass <see cref="MagifyConfig.ClientStateConfig"/> with implemented <see cref="IAppStateProvider"/> to control which state will be saved. <br/>
        /// Magify will also save its state with your data.
        /// </remarks>
        [NotNull, ItemNotNull]
        public static UniTask<SyncStateResult> SaveState(int? weight, CancellationToken cancellationToken)
        {
            ThrowIfMagifyIsNotReady(nameof(SaveState));
            _logger.Log($"Save state with {nameof(weight)}={weight?.ToString() ?? "null"}");
            return MagifyPlatformAPI.SaveState(weight, cancellationToken);
        }

        /// <inheritdoc cref="ThrowIfMagifyIsNotReady"/>
        /// <summary>
        /// Requests restoring a state of the application. You shouldn't call more than one operation in one time.
        /// </summary>
        /// <param name="weight">
        /// A weight that determines how far a player has gone in the game.
        ///     <ul>
        ///         <li> If the weight parameter is specified:
        ///             <ul>
        ///                 <li>if the value <b>is less than</b> the value on the server - progress <b>will</b> be loaded</li>
        ///                 <li>if the value <b>is greater than</b> the value on the server - progress <b>won’t</b> be loaded</li>
        ///             </ul>
        ///         </li>
        ///         <li> If the weight parameter <b>is not specified</b> - progress <b>will</b> be loaded </li>
        ///     </ul>
        /// </param>
        /// <param name="cancellationToken">Pass cancellation token if you want to cancel this operation, for example by timeout.</param>
        /// <returns>
        /// Result of the restore operation.
        /// </returns>
        /// <remarks>
        /// Remember that you should pass <see cref="MagifyConfig.ClientStateConfig"/> with implemented <see cref="IAppStateProvider"/> to handle restored state. <br/>
        /// Magify will also restore its state with your data (but only within one platform).
        /// </remarks>
        [NotNull, ItemNotNull]
        public static UniTask<SyncStateResult> RestoreState(int? weight, CancellationToken cancellationToken)
        {
            ThrowIfMagifyIsNotReady(nameof(RestoreState));
            _logger.Log($"Restore state with {nameof(weight)}={weight?.ToString() ?? "null"}");
            return MagifyPlatformAPI.RestoreState(weight, cancellationToken);
        }

        [CanBeNull]
        public static ICampaign CampaignFor([NotNull] string eventName, CampaignRequestFlags flags, [CanBeNull] Dictionary<string, object> customParams = null)
        {
            var silent = flags.HasFlag(CampaignRequestFlags.Silent);
            _logger.Log($"{nameof(CampaignFor)} called with flags [{flags}] for '{eventName}' (silent={silent}).\nParams:\n{JsonFacade.SerializeObject(customParams, Formatting.Indented)}");
            if (flags.HasFlag(CampaignRequestFlags.ReplaceCampaignTypeWithString) && customParams is { Count: > 0 })
            {
                customParams.ReplaceCampaignTypeWithString();
            }
            return CampaignFor(eventName, customParams, silent);
        }

        public static ICampaign CampaignFor(string eventName, IReadOnlyDictionary<string, object> customParams = null, bool silent = false)
        {
            ThrowIfMagifyIsNotReady(nameof(CampaignFor));
            _logger.Log($"Finding campaign for '{eventName}' (silent={silent}).\nParams:\n{JsonFacade.SerializeObject(customParams, Formatting.Indented)}");
            var campaign = MagifyPlatformAPI?.CampaignFor(eventName, customParams, silent);
            _logger.Log($"Campaign for '{eventName}' (silent={silent}): {(campaign != null ? $"\n{JsonFacade.SerializeObject(campaign, Formatting.Indented)}" : "null")}");
            return campaign;
        }

        public static bool IsCampaignAvailable(string campaignName, string eventName, Dictionary<string, object> customParams = null)
        {
            ThrowIfMagifyIsNotReady(nameof(IsCampaignAvailable));
            _logger.Log($"Checking is campaign '{campaignName}' for '{eventName}' available. Params: \n{JsonFacade.SerializeObject(customParams, Formatting.Indented)}");
            var available = MagifyPlatformAPI.IsCampaignAvailable(campaignName, eventName, customParams);
            _logger.Log($"Campaign '{campaignName}' for '{eventName}' is {(available ? "available" : "not available")}");
            return available;
        }

        public static CampaignImpression LastImpressionFor(CampaignType campaignType, string campaignName, string eventName)
        {
            ThrowIfMagifyIsNotReady(nameof(LastImpressionFor));
            return MagifyPlatformAPI.LastImpressionFor(campaignType, campaignName, eventName);
        }

        // ToDo: make obsolete, replace to SubscribeCampaignUpdates(ICampaign campaign, Action onUpdate)
        public static void SubscribeCampaignUpdates(string campaignName, Action<ICampaign> updateAction)
        {
            MagifyPlatformAPI.SubscribeCampaignUpdates(campaignName, updateAction);
        }

        // ToDo: make obsolete, replace to UnsubscribeCampaignUpdates(string campaignName, Action onUpdate)
        public static void UnsubscribeCampaignUpdates(string campaignName)
        {
            MagifyPlatformAPI.UnsubscribeCampaignUpdates(campaignName);
        }

        public static bool HasProcessedPurchase(string productId)
        {
            ThrowIfMagifyIsNotReady(nameof(HasProcessedPurchase));
            return MagifyPlatformAPI.HasProcessedPurchase(productId);
        }

        /// <summary>
        /// Will return true if creative has default version to show
        /// </summary>
        public static bool HasDefaultCreative(ICampaignWithCreative campaign)
        {
            ThrowIfMagifyIsNotReady(nameof(HasDefaultCreative));
            return campaign.Creative switch
            {
                ImageCreative imageCreative when IsPortrait => Storage.HasDefaultCreative(imageCreative.Portrait),
                ImageCreative imageCreative when !IsPortrait => Storage.HasDefaultCreative(imageCreative.Landscape),
                BundleCreative bundleCreative => Storage.HasDefaultBundle(bundleCreative.Url),
                _ => false
            };
        }

        /// <summary>
        /// Will return true if remote creative has been downloaded
        /// </summary>
        public static bool IsRemoteCreativeDownloaded(ICampaignWithCreative campaign)
        {
            ThrowIfMagifyIsNotReady(nameof(IsRemoteCreativeDownloaded));
            return campaign.Creative switch
            {
                ImageCreative imageCreative when IsPortrait => Storage.ImageIsCached(imageCreative.Portrait),
                ImageCreative imageCreative when !IsPortrait => Storage.ImageIsCached(imageCreative.Landscape),
                BundleCreative bundleCreative => Storage.BundleIsCached(bundleCreative.Url),
                _ => false
            };
        }

        /// <summary>
        /// Will return true if everything has already been downloaded to show the creative.
        /// </summary>
        [Obsolete("Use " + nameof(IsRemoteCreativeDownloaded) + " instead", true)]
        public static bool IsCreativeCached(ICampaignWithCreative campaign)
        {
            throw new NotSupportedException("Use " + nameof(IsRemoteCreativeDownloaded) + " instead");
        }

        #region Tracking

        public static void TrackGdprAccessState(bool accessState)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackGdprAccessState));
            _logger.Log("Track Gdpr Access State");
            MagifyPlatformAPI.TrackGdprAccessState(accessState);
        }

        public static void TrackAppLaunch()
        {
            ThrowIfMagifyIsNotReady(nameof(TrackAppLaunch));
            _logger.Log("Track app launch");
            MagifyPlatformAPI.TrackAppLaunch();
        }

        public static void TrackSubscriptionActivation(bool isTrial, string productId, string price, string currency, string period, string transactionId = null, string purchaseToken = null, string originalTransactionId = null, string receipt = null)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackSubscriptionActivation));
            _logger.Log(
                $"Track subscription activation. {nameof(isTrial)}={isTrial}; {nameof(productId)}={productId}; {nameof(price)}={price}; {nameof(currency)}={currency}; {nameof(period)}={period}; {nameof(transactionId)}={transactionId}; {nameof(purchaseToken)}={purchaseToken}; {nameof(originalTransactionId)}={originalTransactionId}, {nameof(receipt)}={receipt}");
            MagifyPlatformAPI.TrackSubscriptionActivation(isTrial, productId, price, currency, period, transactionId, purchaseToken, originalTransactionId, receipt);
        }

        public static void TrackExternalSubscriptionActivation(bool isTrial, string productId, string price, string currency, string transactionId = null, string purchaseToken = null, string originalTransactionId = null, string receipt = null)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackExternalSubscriptionActivation));
            _logger.Log(
                $"Track external subscription activation. {nameof(isTrial)}={isTrial}; {nameof(productId)}={productId}; {nameof(price)}={price}; {nameof(currency)}={currency}; {nameof(transactionId)}={transactionId}; {nameof(purchaseToken)}={purchaseToken}; {nameof(originalTransactionId)}={originalTransactionId}, {nameof(receipt)}={receipt}");
            MagifyPlatformAPI.TrackExternalSubscriptionActivation(isTrial, productId, price, currency, transactionId, purchaseToken, originalTransactionId, receipt);
        }

        public static void TrackInAppFor(string productId, string price, string currency, string transactionId = null, string purchaseToken = null, string originalTransactionId = null, string receipt = null)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackInAppFor));
            _logger.Log($"Track inapp. {nameof(productId)}={productId}; {nameof(price)}={price}; {nameof(currency)}={currency}; {nameof(transactionId)}={transactionId}; {nameof(purchaseToken)}={purchaseToken}; {nameof(originalTransactionId)}={originalTransactionId}, {nameof(receipt)}={receipt}");
            MagifyPlatformAPI.TrackInAppFor(productId, price, currency, transactionId, purchaseToken, originalTransactionId, receipt);
        }

        public static void TrackExternalInAppFor(string productId, string price, string currency, string transactionId = null, string purchaseToken = null, string originalTransactionId = null, string receipt = null)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackExternalInAppFor));
            _logger.Log($"Track external inapp. {nameof(productId)}={productId}; {nameof(price)}={price}; {nameof(currency)}={currency}; {nameof(transactionId)}={transactionId}; {nameof(purchaseToken)}={purchaseToken}; {nameof(originalTransactionId)}={originalTransactionId}, {nameof(receipt)}={receipt}");
            MagifyPlatformAPI.TrackExternalInAppFor(productId, price, currency, transactionId, purchaseToken, originalTransactionId, receipt);
        }

        public static void TrackRestoredInAppFor(string productId)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackRestoredInAppFor));
            _logger.Log($"Track restored inapp. {nameof(productId)}={productId}");
            MagifyPlatformAPI.TrackRestoredInAppFor(productId);
        }

        public static void TrackRestoredSubscription(string productId)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackRestoredSubscription));
            _logger.Log($"Track restored subscription. {nameof(productId)}={productId}");
            MagifyPlatformAPI.TrackRestoredSubscription(productId);
        }

        public static void TrackProductClickFor(CampaignType campaignType, string productId)
        {
            switch (productId)
            {
                case ProductDef.FakeProductId:
                    TrackClickFor(campaignType);
                    break;
                default:
                    ThrowIfMagifyIsNotReady(nameof(TrackProductClickFor));
                    _logger.Log($"Track product click. {nameof(campaignType)}={campaignType}; {nameof(productId)}={productId}");
                    MagifyPlatformAPI.TrackProductClickFor(campaignType, productId);
                    break;
            }
        }

        public static void TrackAdsProductClickFor(CampaignType campaignType, string productId)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackAdsProductClickFor));
            _logger.Log($"Track ads product click. {nameof(campaignType)}={campaignType}; {nameof(productId)}={productId}");
            MagifyPlatformAPI?.TrackAdsProductClickFor(campaignType, productId);
        }

        public static void TrackClickFor(CampaignType campaignType)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackClickFor));
            _logger.Log($"Track click. {nameof(campaignType)}={campaignType}");
            MagifyPlatformAPI.TrackClickFor(campaignType);
        }

        public static void TrackAdsClickFor(CampaignType campaignType)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackAdsClickFor));
            _logger.Log($"Track ads click. {nameof(campaignType)}={campaignType}");
            MagifyPlatformAPI?.TrackAdsClickFor(campaignType);
        }

        public static void TrackImpressionFailFor(CampaignType campaignType, string reason)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackImpressionFailFor));
            _logger.Log($"Track fail. {nameof(campaignType)}={campaignType}; {nameof(reason)}={reason}");
            MagifyPlatformAPI.TrackImpressionFailFor(campaignType, reason);
        }

        public static void TrackAdsImpression(CampaignType campaignType, IAdsImpression impression)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackAdsImpression));
            _logger.Log($"Track ads impression. {nameof(campaignType)}={campaignType}; {nameof(impression)}={JsonFacade.SerializeObject(impression)}");
            MagifyPlatformAPI.TrackAdsImpression(campaignType, impression);
        }

        public static void TrackImpression(CampaignType campaignType)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackImpression));
            _logger.Log($"Track impression. {nameof(campaignType)}={campaignType}");
            MagifyPlatformAPI.TrackImpression(campaignType);
        }

        public static void TrackParentCampaignImpression(CampaignType campaignType)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackParentCampaignImpression));
            _logger.Log($"Track parent campaign impression. {nameof(campaignType)}={campaignType}");
            MagifyPlatformAPI.TrackParentCampaignImpression(campaignType);
        }

        public static void TrackProductsImpression(CampaignType campaignType, List<string> productIds)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackProductsImpression));
            productIds.RemoveAll(c => c == ProductDef.FakeProductId);
            if (productIds.Count == 0) return;
            _logger.Log($"Track products impression. {nameof(campaignType)}={campaignType}; {nameof(productIds)}={JsonFacade.SerializeObject(productIds)}");
            MagifyPlatformAPI.TrackProductsImpression(campaignType, productIds);
        }

        public static void TrackCustomEvent(string eventName, IReadOnlyDictionary<string, object> customParams = null)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackCustomEvent));
            _logger.Log($"Track custom event. {nameof(eventName)}={eventName}; {nameof(customParams)}={JsonFacade.SerializeObject(customParams)}");
            MagifyPlatformAPI.TrackCustomEvent(eventName, customParams);
        }

        public static void TrackIncomeTransaction(string source, List<BonusInfo> bonuses, ProductInfo product = null)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackIncomeTransaction));
            _logger.Log($"Track income transaction. {nameof(source)}={source}; {nameof(bonuses)}={JsonFacade.SerializeObject(bonuses)}; {nameof(product)}={JsonFacade.SerializeObject(product)}");
            MagifyPlatformAPI.TrackIncomeTransaction(source, bonuses, product);
        }

        public static void TrackExpenseTransaction(List<BonusInfo> bonuses)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackExpenseTransaction));
            _logger.Log($"Track expense transaction. {nameof(bonuses)}={JsonFacade.SerializeObject(bonuses)}");
            MagifyPlatformAPI.TrackExpenseTransaction(bonuses);
        }

        public static void TrackCorrectionTransaction(List<BonusInfo> bonuses)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackCorrectionTransaction));
            _logger.Log($"Track correction transaction. {nameof(bonuses)}={JsonFacade.SerializeObject(bonuses)}");
            MagifyPlatformAPI.TrackCorrectionTransaction(bonuses);
        }

        public static void TrackRewardGranted(string productId)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackRewardGranted));
            _logger.Log($"Track reward granted. {nameof(productId)}={productId}");
            MagifyPlatformAPI.TrackRewardGranted(productId);
        }

        public static void TrackFreeBonusGranted(string productId)
        {
            ThrowIfMagifyIsNotReady(nameof(TrackFreeBonusGranted));
            _logger.Log($"Track free bonus granted. {nameof(productId)}={productId}");
            MagifyPlatformAPI.TrackFreeBonusGranted(productId);
        }

        public static void TrackOrdinaryProductUsed(string productId)
        {
            if (productId is ProductDef.FakeProductId) return;
            ThrowIfMagifyIsNotReady(nameof(TrackOrdinaryProductUsed));
            _logger.Log($"Track ordinary product used. {nameof(productId)}={productId}");
            MagifyPlatformAPI.TrackOrdinaryProductUsed(productId);
        }

        #endregion

        #region Debugging

        /// <exception cref="MagifyNotInitializedException">
        /// If you try to call this method before magify has been initialized.
        /// </exception>
        internal static void ThrowIfMagifyIsNotReady(string caller)
        {
            if (MagifyPlatformAPI == null)
            {
                throw new MagifyNotInitializedException(caller);
            }
        }

        #endregion
    }
}